/* gxxfvfl.c */
/* eXternal Font implementation for VFlib --- Drawing routines */
/*
Akira Tanaka <akr@jaist.ac.jp> created for gs5.10-vflib-1.4. 1998/05/12
  derived from zkfvflib.c
*/
/*
 * zkfvflib.c --- Kanji font operator for VFlib.
 * Copyright (C) ASAYAMA Kazunori  1994.
 * Akira Tanaka <akr@jaist.ac.jp> modified for gs4.03/VFlib support. 1996/10/17
 */
#include "math_.h"
#include "memory_.h"
#include "ghost.h"
#include "gp.h"
#include "oper.h"
#include "errors.h"
#include "gspath.h"
#include "gscoord.h"
#include "gsstate.h"
#include "igstate.h"

#include "gxdevice.h"
#include "gxxfont.h"
#include "gxxfvf.h"

/* Including the header file for VFlib. */
#include "VF.h"
#ifdef __MSDOS__
#  include "dos_.h"
#endif

#define D_X_CL	(OUTLINE_SIZE * 0.5 + OUTLINE_OFFSET)
#define D_Y_CL	(OUTLINE_SIZE * 0.5 + OUTLINE_OFFSET)
#define	D_Y_BL	(OUTLINE_SIZE * 0.88 + OUTLINE_OFFSET)

#define B_X_CL  VF_CHAR_CENTER_X	/* horizontal center in BuildChar */
#define B_Y_CL  VF_CHAR_CENTER_Y	/* vertical center in BuildChar */
#define B_Y_BL  VF_CHAR_BASE	/* base line in BuildChar */

/* Drawing routine. */

/* data coordination -> character coordination */
/*#define Y_FACTOR (-(floatp)1000 / (floatp)OUTLINE_SIZE)*/
#define Y_FACTOR ((floatp)(B_Y_CL - B_Y_BL) / (floatp)(D_Y_CL - D_Y_BL))
#define X_FACTOR ((floatp)(Y_FACTOR > 0 ? Y_FACTOR : - Y_FACTOR))
#define map_x(x) ((floatp)B_X_CL + ((floatp)(x) - (floatp)D_X_CL) * X_FACTOR)
#define map_y(y) ((floatp)B_Y_CL + ((floatp)(y) - (floatp)D_Y_CL) * Y_FACTOR - B_Y_BL)

/* drawing */
#define moveto(x, y)    gs_moveto(igs, map_x(x), map_y(y))
#define lineto(x, y)    gs_lineto(igs, map_x(x), map_y(y))
#define curveto(x1, y1, x2, y2, x3, y3) \
  gs_curveto(igs, map_x(x1), map_y(y1), \
             map_x(x2), map_y(y2), map_x(x3), map_y(y3))
#define closepath() gs_closepath(igs)

#define	vf_is_clock(c)     (((c) & VFD_CWCURV) == VFD_CWCURV)
#define	vf_is_a_clock(c)   (((c) & VFD_CCWCURV) == VFD_CCWCURV)
#define	vf_is_bos(c)       (vf_is_clock(c) || vf_is_a_clock(c))
#define	vf_is_token(c)	   (((c) & VFD_TOKEN) == VFD_TOKEN)

#define vf_get_xy(ptr, x, y) \
  if (vf_is_bos(ptr)) return e_undefinedresult; \
  else (x = VFD_GET_X(ptr), y = VFD_GET_Y(ptr))
#define VFD_GMASK (VFD_LINE | VFD_ARC | VFD_BEZ)

private double
atan360(double y, double x)
{
  double ang;

  if((ang = atan2(y, x) * radians_to_degrees) < 0)
    ang += 360.0;
  return ang;
}

/* darc3 */
private int
darc3(i_ctx_t *i_ctx_p, floatp x1, floatp y1, floatp x2,
      floatp y2, floatp x3, floatp y3)
{
  int code;
  floatp dx1, dy1, dx3, dy3, cx, cy, z, r, ang1, ang2, ang3, ang;

  x1 = map_x(x1);		y1 = map_y(y1);
  x2 = map_x(x2);		y2 = map_y(y2);
  x3 = map_x(x3);		y3 = map_y(y3);

  dx1 = x1 - x2;		dy1 = y1 - y2;
  dx3 = x3 - x2;		dy3 = y3 - y2;

  if((z = dx1*dy3 - dx3*dy1) == 0) {
    if((dx1 == dx3) && (dy1 == dy3)) {
      cx = dx1 / 2.0;
      cy = dy1 / 2.0;
      r = sqrt(cx*cx + cy*cy);
      ang = atan360(dy1 - cy, dx1 - cx);
      return gs_arc(igs, cx + x2, cy + y2, r, ang, ang+360);
    }
    else {
      if((code = gs_lineto(igs, x1, y1)) < 0)
	return code;
      return gs_lineto(igs, x3, y3);
    }
  }

  cx = ((dx1*dx1 + dy1*dy1)*dy3 - (dx3*dx3 + dy3*dy3)*dy1) / z / 2.0;
  cy = - ((dx1*dx1 + dy1*dy1)*dx3 - (dx3*dx3 + dy3*dy3)*dx1) / z / 2.0;
  r = sqrt(cx*cx + cy*cy);

  ang1 = atan360(dy1 - cy, dx1 - cx);
  ang2 = atan360(-cy, -cx);
  ang3 = atan360(dy3 - cy, dx3 - cx);

  if(ang1 == ang3)
    return gs_arc(igs, cx + x2, cy + y2, r, 0.0, 360.0);
  else if(ang1 < ang3) {
    if(ang1 <= ang2 && ang2 <= ang3)
      return gs_arc(igs, cx + x2, cy + y2, r, ang1, ang3);
    else
      return gs_arcn(igs, cx + x2, cy + y2, r, ang1, ang3);
  }
  else {
    if(ang3 <= ang2 && ang2 <= ang1)
      return gs_arcn(igs, cx + x2, cy + y2, r, ang1, ang3);
    else
      return gs_arc(igs, cx + x2, cy + y2, r, ang1, ang3);
  }
}

int
vf_draw_font(i_ctx_t *i_ctx_p, const byte *font_name, int len, int jis)
{
  int code;
  int font_fd;
  long *ptr,*outline;
  long token,next_token;
  int x0,x1,x2,x3,y0,y1,y2,y3;

  font_fd = vf_open_font(font_name, len, 1);
  if (font_fd < 0)
    return font_fd;

  /* Get out-line data. */
  outline = (long *)VF_GetOutline(jis,font_fd);

  if (outline == NULL)
    return 0;

  ptr = outline + 2;

  token = *ptr++;
  while(token) {
    /* The beginning of segment */
    if(!vf_is_bos(token))
      return e_undefinedresult;
    token &= VFD_GMASK;

    vf_get_xy(*ptr, x0, y0);
    ptr ++;
    x3 = x0; y3 = y0;
    code = moveto((floatp)x0, (floatp)y0);
    if (code<0) return code;

    while(!vf_is_bos(token) && token) {
      switch(token & VFD_GMASK) {
      case VFD_LINE:
        break;
      case VFD_ARC:
        x1 = x3; y1 = y3;
        vf_get_xy(*ptr, x2, y2);
        ptr ++;
        break;
      case VFD_BEZ:
        vf_get_xy(*ptr, x1, y1);
        ptr ++;
        vf_get_xy(*ptr, x2, y2);
        ptr ++;
        break;
      default:
        return e_undefinedresult;
      }
      if (vf_is_token(*ptr) || *ptr == 0) {
        next_token = *ptr++;
        if (vf_is_bos(next_token) || next_token == 0) {
          x3 = x0; y3 = y0;
        }
        else {
          vf_get_xy(*ptr, x3, y3);
          ptr ++;
        }
      }
      else {
        next_token = token;
        vf_get_xy(*ptr, x3, y3);
        ptr ++;
      }
      switch(token & VFD_GMASK) {
      case VFD_LINE:
        code = lineto((floatp)x3, (floatp)y3);
        break;
      case VFD_ARC:
        code = darc3(i_ctx_p,(floatp)x1, (floatp)y1,
                     (floatp)x2, (floatp)y2,
                     (floatp)x3,  (floatp)y3);
        break;
      case VFD_BEZ:
        code = curveto((floatp)x1, (floatp)y1,
                       (floatp)x2, (floatp)y2,
                       (floatp)x3,  (floatp)y3);
        break;
      default:
        code = e_undefinedresult;
        break;
      }
      if(code<0) return code;

      token = next_token;
    }
    code = closepath();
    if (code<0) return code;
  }
  /* free buffer area. */
  VF_FreeOutline((char *)outline,font_fd);

  return 0;
}
