/* gxxfvfl.c */
/* eXternal Font implementation for VFlib --- Painting routines */
/*
Akira Tanaka <akr@jaist.ac.jp> created for gs5.10-vflib-1.4. 1998/05/13
*/

#include "VF.h"
#include "math_.h"

#include "gx.h"
#include "gxdevice.h"
#include "gxxfont.h"
#include "gxxfvf.h"

static byte minbit[256] = {
  0, 7, 6, 6, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4,
  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
  2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
  2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

static byte maxbit[256] = {
  0, 7, 6, 7, 5, 7, 6, 7, 4, 7, 6, 7, 5, 7, 6, 7,
  3, 7, 6, 7, 5, 7, 6, 7, 4, 7, 6, 7, 5, 7, 6, 7,
  2, 7, 6, 7, 5, 7, 6, 7, 4, 7, 6, 7, 5, 7, 6, 7,
  3, 7, 6, 7, 5, 7, 6, 7, 4, 7, 6, 7, 5, 7, 6, 7,
  1, 7, 6, 7, 5, 7, 6, 7, 4, 7, 6, 7, 5, 7, 6, 7,
  3, 7, 6, 7, 5, 7, 6, 7, 4, 7, 6, 7, 5, 7, 6, 7,
  2, 7, 6, 7, 5, 7, 6, 7, 4, 7, 6, 7, 5, 7, 6, 7,
  3, 7, 6, 7, 5, 7, 6, 7, 4, 7, 6, 7, 5, 7, 6, 7,
  0, 7, 6, 7, 5, 7, 6, 7, 4, 7, 6, 7, 5, 7, 6, 7,
  3, 7, 6, 7, 5, 7, 6, 7, 4, 7, 6, 7, 5, 7, 6, 7,
  2, 7, 6, 7, 5, 7, 6, 7, 4, 7, 6, 7, 5, 7, 6, 7,
  3, 7, 6, 7, 5, 7, 6, 7, 4, 7, 6, 7, 5, 7, 6, 7,
  1, 7, 6, 7, 5, 7, 6, 7, 4, 7, 6, 7, 5, 7, 6, 7,
  3, 7, 6, 7, 5, 7, 6, 7, 4, 7, 6, 7, 5, 7, 6, 7,
  2, 7, 6, 7, 5, 7, 6, 7, 4, 7, 6, 7, 5, 7, 6, 7,
  3, 7, 6, 7, 5, 7, 6, 7, 4, 7, 6, 7, 5, 7, 6, 7
};

private void
vtransform(
  int font_fd,
  int jis_code,
  int wmode,
  gs_matrix *pmat)
{
#define METRICS_RASTER 5
#define METRICS_WIDTH (METRICS_RASTER*8)
#define METRICS_HEIGHT METRICS_WIDTH
  static byte buf[METRICS_RASTER * METRICS_HEIGHT];
  int x, y, b;
  gs_rect bbox;

  if(!wmode)
  {
    gs_make_identity(pmat);
    return;
  }

  memset(buf, 0, METRICS_RASTER * METRICS_HEIGHT);
  VF_GetBitmap(jis_code, font_fd,
    METRICS_WIDTH, METRICS_HEIGHT, METRICS_RASTER, 0, buf);

  for(y = 0; y < METRICS_HEIGHT; y++)
    for(x = 0; x < METRICS_RASTER; x++)
      if((b = buf[y * METRICS_RASTER + x]) != 0)
      {
        bbox.p.x = x * 8 + minbit[b];
        bbox.p.y = y;
        bbox.q.x = x * 8 + maxbit[b];
        bbox.q.y = y;
        goto found_pixel;
      }
  bbox.p.x = bbox.p.y = 0;
  bbox.q.x = bbox.q.y = 0;
  goto found_bbox;

  for(y = 0; y < METRICS_HEIGHT; y++)
    for(x = 0; x < METRICS_RASTER; x++)
      if((b = buf[y * METRICS_RASTER + x]) != 0)
      {
        if(bbox.p.x > x * 8 + minbit[b]) bbox.p.x = x * 8 + minbit[b];
        if(bbox.p.y > y) bbox.p.y = y;
        if(bbox.q.x < x * 8 + maxbit[b]) bbox.q.x = x * 8 + maxbit[b];
        if(bbox.q.y < y) bbox.q.y = y;
        found_pixel:;
      }

  found_bbox:
  {
    int y = bbox.p.y;
    bbox.p.y = METRICS_HEIGHT - bbox.q.y - 1;
    bbox.q.y = METRICS_HEIGHT - y - 1;
  }
  bbox.q.x += 1;
  bbox.q.y += 1;

  bbox.p.x = bbox.p.x * 1000 / METRICS_WIDTH;
  bbox.p.y = bbox.p.y * 1000 / METRICS_HEIGHT;
  bbox.q.x = bbox.q.x * 1000 / METRICS_WIDTH;
  bbox.q.y = bbox.q.y * 1000 / METRICS_HEIGHT;
  vf_vmatrix(jis_code, &bbox, pmat);
}

/* integer matrix routines */

typedef struct { int xx, yx, xy, yy, tx, ty; } imatrix;
typedef struct { gs_int_point o, u, v; } ispace;

#define sign3(r) ((r) < 0 ? -1 : (r) == 0 ? 0 : 1)
#define sign2(r) ((r) < 0 ? -1 : 1)
#define round(r) (ceil((r) - 0.5))

private int
imatrix_multiply(
  imatrix *im1,
  imatrix *im2,
  imatrix *im3)
{
  imatrix ima = *im1;
  imatrix imb = *im2;
  im3->xx = ima.xx * imb.xx + ima.xy * imb.yx;
  im3->xy = ima.xx * imb.xy + ima.xy * imb.yy;
  im3->yx = ima.yx * imb.xx + ima.yy * imb.yx;
  im3->yy = ima.yx * imb.xy + ima.yy * imb.yy;
  im3->tx = ima.tx * imb.xx + ima.ty * imb.yx + imb.tx;
  im3->ty = ima.tx * imb.xy + ima.ty * imb.yy + imb.ty;
  return 0;
}

private int
int_point_transform(
  int x,
  int y,
  imatrix *m,
  gs_int_point *p)
{
  p->x = m->xx * x + m->yx * y + m->tx;
  p->y = m->xy * x + m->yy * y + m->ty;
  return 0;
}

private int
imatrix_invert(
  imatrix *im1,
  imatrix *im2)
{
  imatrix im = *im1;
  if(abs(im.xx) == 1 && im.xy == 0 && im.yx == 0 && abs(im.yy) == 1)
  {
    im2->xx = im.xx;
    im2->xy = 0;
    im2->yx = 0;
    im2->yy = im.yy;
    im2->tx = -im.xx * im.tx;
    im2->ty = -im.yy * im.ty;
  }
  else if(im.xx == 0 && abs(im.xy) == 1 && abs(im.yx) == 1 && im.yy == 0)
  {
    im2->xx = 0;
    im2->xy = im.yx;
    im2->yx = im.xy;
    im2->yy = 0;
    im2->tx = -im.xy * im.ty;
    im2->ty = -im.yx * im.tx;
  }
  else
    return -1;
  return 0;
}

private int
imatrix_make_transform(
  ispace *space,
  imatrix *im,
  int *width,
  int *height)
{
  int ux = space->u.x - space->o.x;
  int uy = space->u.y - space->o.y;
  int vx = space->v.x - space->o.x;
  int vy = space->v.y - space->o.y;

  if((ux == 0) == (uy == 0)) return -1;
  if((vx == 0) == (vy == 0)) return -1;
  if((ux == 0) == (vx == 0)) return -1;

  im->xx = sign3(ux);
  im->xy = sign3(uy);
  im->yx = sign3(vx);
  im->yy = sign3(vy);
  im->tx = space->o.x;
  im->ty = space->o.y;
  if(width) *width = abs(ux + uy);
  if(height) *height = abs(vx + vy);
  return 0;
}

private int
ispace_transform(
  ispace *space0,
  gs_matrix *pmat,
  ispace *space1)
{
  gs_point p;
  gs_point_transform(space0->o.x, space0->o.y, pmat, &p);
  space1->o.x = round(p.x); space1->o.y = round(p.y);
  gs_point_transform(space0->u.x, space0->u.y, pmat, &p);
  space1->u.x = round(p.x); space1->u.y = round(p.y);
  gs_point_transform(space0->v.x, space0->v.y, pmat, &p);
  space1->v.x = round(p.x); space1->v.y = round(p.y);
  return 0;
}

private int
ispace_bbox(
  ispace *s,
  gs_int_rect *b)
{
  int i;
  int minx, maxx, miny, maxy;
/*
  gs_int_point p[] = {
    { s->u.x, s->u.y },
    { s->v.x, s->v.y },
    { s->u.x + s->v.x - s->o.x, s->u.y + s->v.y - s->o.y }
  };
*/
  gs_int_point p[3];
  p[0].x=s->u.x;
  p[0].y=s->u.y;
  p[1].x=s->v.x;
  p[1].y=s->v.y;
  p[2].x=s->u.x + s->v.x - s->o.x;
  p[2].y=s->u.y + s->v.y - s->o.y;

  minx = maxx = s->o.x;
  miny = maxy = s->o.y;
  for(i = 0; i < sizeof(p)/sizeof(*p); i++)
  {
    if(p[i].x < minx) minx = p[i].x;
    if(p[i].y < miny) miny = p[i].y;
    if(maxx < p[i].x) maxx = p[i].x;
    if(maxy < p[i].y) maxy = p[i].y;
  }
  b->p.x = minx;
  b->p.y = miny;
  b->q.x = maxx;
  b->q.y = maxy;
  return 0;
}

private int
ispace_instansiate_gs(
  ispace *s,
  imatrix *m,
  int *w,
  int *h,
  int *raster,
  int *size)
{
  int code;
  gs_int_rect b;
  ispace ss;

  ispace_bbox(s, &b);
  {
/*
     ispace s = { { b.p.x, b.p.y }, { b.q.x, b.p.y }, { b.p.x, b.q.y } };
*/
    ss.o.x = b.p.x;
    ss.o.y = b.p.y;
    ss.u.x = b.q.x;
    ss.u.y = b.p.y;
    ss.v.x = b.p.x;
    ss.v.y = b.q.y;
/*
    if((code = imatrix_make_transform(&s, m, NULL, NULL)) < 0)
*/
    if((code = imatrix_make_transform(&ss, m, NULL, NULL)) < 0)
      return code;
  }
  *w = abs(b.q.x - b.p.x);
  *h = abs(b.q.y - b.p.y);
  *raster = round_up(*w, align_bitmap_mod * 8) >> 3;
  *size = *raster * *h;
  return 0;
}

private int
ispace_instansiate_vf(
  ispace *ps,
  imatrix *m,
  int *w,
  int *h,
  int *raster,
  int *size)
{
  int code;
  gs_int_rect b;
  gs_int_point t;
  ispace s = *ps;

  s.u.x += s.v.x - s.o.x;
  s.u.y += s.v.y - s.o.y;
  t = s.v; s.v = s.o; s.o = t;

  ispace_bbox(&s, &b);
  if((code = imatrix_make_transform(&s, m, w, h)) < 0)
    return code;
  *raster = (*w + 7)/8;
  *size = *raster * *h;
  return 0;
}

int
vf_paint_font(
  int font_fd,
  gs_matrix *pmat,
  int jis_code,
  int char_wmode,
  int paint_wmode,
  byte **pbuf,
  int *praster,
  gs_int_rect *pbbox,
  gs_point *pwidth)
{
  static int clen = 0;
  static byte *cbuf;

  static int rlen = 0;
  static byte *rbuf;
  static int raster;
  static gs_int_rect bbox;
  static gs_point width;

  static int cache_font_fd = -1;
  static gs_matrix cache_mat;
  static int cache_jis_code;
  static int cache_char_wmode;
  static int cache_paint_wmode;

  int rh, rw, rr, rs;
  int ch, cw, cr, cs;
  ispace sp = {
    {    0,    0 - VF_CHAR_BASE },
    { 1000,    0 - VF_CHAR_BASE },
    {    0, 1000 - VF_CHAR_BASE }
  };
  ispace rp, cp;
  gs_matrix vmat;
  imatrix imat;

  if(0 <= cache_font_fd &&      /* cache is valid */
    cache_font_fd == font_fd &&
    cache_mat.xx == pmat->xx &&
    cache_mat.xy == pmat->xy &&
    cache_mat.yx == pmat->yx &&
    cache_mat.yy == pmat->yy &&
    cache_mat.tx == pmat->tx &&
    cache_mat.ty == pmat->ty &&
    cache_jis_code == jis_code &&
    cache_char_wmode == char_wmode &&
    cache_paint_wmode == paint_wmode)
  {
    *pbuf = rbuf;
    *praster = raster;
    *pbbox = bbox;
    if(pwidth) *pwidth = width;
    return 0;
  }
  cache_font_fd = -1; /* invalidate cache */

#if 0
  fprintf(stdout, "paint_font %04x %s %s\n", jis_code, char_wmode?"V":"H", paint_wmode?"V":"H");
  fprintf(stdout, "paint_font pmat:[%f %f %f %f %f %f]\n", pmat->xx, pmat->xy, pmat->yx, pmat->yy, pmat->tx, pmat->ty);
#endif

  vtransform(font_fd, jis_code, char_wmode, &vmat);
  ispace_transform(&sp, pmat, &rp);

#if 0
  fprintf(stdout, "paint_font vmat:[%f %f %f %f %f %f]\n", vmat.xx, vmat.xy, vmat.yx, vmat.yy, vmat.tx, vmat.ty);
  fprintf(stdout, "paint_font rp o[%d %d] u[%d %d] v[%d %d]\n", rp.o.x, rp.o.y, rp.u.x, rp.u.y, rp.v.x, rp.v.y);
#endif

  {
    gs_matrix vp;
    gs_matrix_multiply(&vmat, pmat, &vp);
    ispace_transform(&sp, &vp, &cp);
  }
#if 0
  fprintf(stdout, "paint_font cp o[%d %d] u[%d %d] v[%d %d]\n", cp.o.x, cp.o.y, cp.u.x, cp.u.y, cp.v.x, cp.v.y);
#endif

  {
    imatrix rmat, cmat;
    ispace_instansiate_gs(&rp, &rmat, &rw, &rh, &rr, &rs);
    ispace_instansiate_vf(&cp, &cmat, &cw, &ch, &cr, &cs);
    imatrix_invert(&cmat, &cmat);
    imatrix_multiply(&rmat, &cmat, &imat);

#if 0
    fprintf(stdout, "paint_font rw:%d rh:%d rmat:[%d %d %d %d %d %d]\n", rw, rh, rmat.xx, rmat.xy, rmat.yx, rmat.yy, rmat.tx, rmat.ty);
    fprintf(stdout, "paint_font cw:%d ch:%d cmat:[%d %d %d %d %d %d]\n", cw, ch, cmat.xx, cmat.xy, cmat.yx, cmat.yy, cmat.tx, cmat.ty);
    fprintf(stdout, "paint_font imat:[%d %d %d %d %d %d]\n", imat.xx, imat.xy, imat.yx, imat.yy, imat.tx, imat.ty);
#endif
  }

  if(rlen == 0)
  { rlen = rs; rbuf = (byte *)malloc(rs); if(rbuf == NULL) return -1; }
  else if(rlen < rs)
  { rlen = rs; rbuf = (byte *)realloc(rbuf, rlen); if(rbuf == NULL) return -1; }
  memset(rbuf, 0, rs);

  if(clen == 0)
  { clen = cs; cbuf = (byte *)malloc(cs); if(cbuf == NULL) return -1; }
  else if(clen < cs)
  { clen = cs; cbuf = (byte *)realloc(cbuf, clen); if(cbuf == NULL) return -1; }
  memset(cbuf, 0, cs);

  VF_GetBitmap(jis_code, font_fd, cw, ch, cr, 0, cbuf);

  {
    int rx, ry;
    for(ry = 0; ry < rh; ry++)
      for(rx = 0; rx < rw; rx++)
      {
        gs_int_point c;
        int_point_transform(rx, ry, &imat, &c);
        if(0 <= c.x && c.x < cw && 0 <= c.y && c.y < ch)
        {
#if 0
          fprintf(stdout, "paint_font ry:%d rx:%d cx:%d cy:%d\n", ry, rx, c.x, c.y);
#endif
          if(cbuf[cr * c.y + (c.x >> 3)] & (0x80 >> (c.x & 7)))
            rbuf[rr * ry + (rx >> 3)] |= (0x80 >> (rx & 7));
        }
      }
  }

  raster = rr;
  ispace_bbox(&rp, &bbox);

  if(paint_wmode)
  {
    gs_point v = { VF_CHAR_CENTER_X, VF_CHAR_HEIGHT - VF_CHAR_BASE };
    gs_distance_transform(v.x, v.y, pmat, &v);
    bbox.p.x -= v.x;
    bbox.p.y -= v.y;
    bbox.q.x -= v.x;
    bbox.q.y -= v.y;
  }

  if(paint_wmode)
  {
    gs_point w = { 0, -VF_CHAR_HEIGHT };
    gs_distance_transform(w.x, w.y, pmat, &width);
  }
  else
  {
    gs_point w = { VF_CHAR_WIDTH, 0 };
    gs_point_transform(w.x, w.y, pmat, &width);
  }

  cache_font_fd = font_fd;
  cache_mat.xx = pmat->xx;
  cache_mat.xy = pmat->xy;
  cache_mat.yx = pmat->yx;
  cache_mat.yy = pmat->yy;
  cache_mat.tx = pmat->tx;
  cache_mat.ty = pmat->ty;
  cache_jis_code = jis_code;
  cache_char_wmode = char_wmode;
  cache_paint_wmode = paint_wmode;

  *pbuf = rbuf;
  *pbbox = bbox;
  *praster = raster;
  if(pwidth) *pwidth = width;
  return 0;
}
