/*
 * zkfvflib.c --- Kanji font operator for VFlib.
 * Copyright (C) ASAYAMA Kazunori  1994.
 * Akira Tanaka <akr@jaist.ac.jp> modified for gs4.03/VFlib support. 1996/10/17
 * Akira Tanaka <akr@jaist.ac.jp> modified for gs5.10-vflib-1.3. 1998/05/12
 * Akira Tanaka <akr@jaist.ac.jp> modified for gs5.10-vflib-1.4. 1998/05/12
 */

#include "ghost.h"
#include "oper.h"
#include "opdef.h"
#include "gspath.h"
#include "gsmatrix.h"
#include "gscoord.h"
#include "igstate.h"
#include "gxcindex.h"

#include "gsuid.h"
#include "gxxfont.h"
#include "kfutil.h"
#include "gxxfvf.h"

/* zkfvflib */
private int
zkfVFlib(i_ctx_t *i_ctx_p)
{
  os_ptr op = osp;
  char *font_name;
  int len;
  int code, jis_code, wmode;
  
  check_type(op[-2], t_integer);                /* JIS Code */
  check_type(op[-1], t_integer);                /* WMode */
  check_type(op[0], t_string);          /* Font Name */ 
  
  font_name = op->value.bytes;
  len = r_size(op);

  jis_code = op[-2].value.intval;
  wmode = op[-1].value.intval;
  
  if(wmode && kf_is_vchar(jis_code)) {
    /* vertical kanji character */
    gs_matrix smat, vmat;
    gs_rect bbox;
    
    if((code = gs_gsave(igs)) < 0)
      return code;
    if((code = vf_draw_font(i_ctx_p, font_name, len, jis_code)) < 0)
      return code;
    if((code = gs_flattenpath(igs)) < 0)
      return code;
    if((code = gs_pathbbox(igs, &bbox)) < 0)
      return code;
    if((code = gs_grestore(igs)) < 0)
      return code;
    
    gs_currentmatrix(igs, &smat);
    vf_vmatrix(jis_code, &bbox, &vmat);

    if((code = gs_concat(igs, &vmat)) < 0)
      return code;
    if((code = vf_draw_font(i_ctx_p, font_name, len, jis_code)) < 0)
      return code;
    gs_setmatrix(igs, &smat);
  }
  else if((code = vf_draw_font(i_ctx_p, font_name, len, jis_code)) < 0) {
    return code;
  }
  
  pop(3);
  return 0;
}

/* -------- Initialization procedure -------- */

const op_def zkfVFlib_op_defs[] = {
  {"3kfVFlib", zkfVFlib},
  op_def_end(0)
};
